package main

import (
	"fmt"
	"time"
	"math/rand"
	"strings"
	"os"
	"bufio"
	"strconv"
)


//////////////////////////////////////// MAIN ////////////////////////////////////

func main() {
	fmt.Println("----------------------------------------------------------------------")
    fmt.Println("Welcome to the game of BlackJack! Your starting money: $100. Have fun!")
    fmt.Println("----------------------------------------------------------------------")
    rand.Seed(time.Now().UnixNano())
    run()
    
}

////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////// PLAYER ///////////////////////////////////

//The struct representing the player - dealer or player
type Player struct {
	cards []Card
	state int
	optimalValue int
}

//A player that has money inherited from player
type BettingPlayer struct {
	money int
	Player
}

//stand operation of the player
func (player *Player) stand(values []int) {
	_player.state = STAND
	_player.optimalValue = getMaximum(values)
}

//player's hit operation
func (player *Player) hit() {
	player.cards = append(player.cards,_deck.getCard())
}

//writes out the player's cards and values
func (player Player) writeCards() ([]int) {
	for i, _ := range(player.cards) {
		fmt.Print( player.cards[i].name, " of ",  player.cards[i].color, ", ")
	}
	fmt.Println(" ")
	fmt.Println("Possible values: ")
	values := player.calculatePossibleValues()
	for _, c := range(values) {
		fmt.Print(c, " ")
	}
	fmt.Println(" ")
	return values
}

//calculates the possible values of the current cards of the player
func (player Player) calculatePossibleValues() ([]int) {
	possibleValues := make([]int,1)
	possibleValues[0] = 0
	for _, c := range(player.cards) {
		if c.name != "Ace" {
			for i, _ := range(possibleValues) {
				possibleValues[i] += c.value
			}
		} else {
			length := len(possibleValues)
			for i := 0; i<length; i++ {
				possibleValues = append(possibleValues,possibleValues[i]+c.alternateValue)
				possibleValues[i] += c.value
			}
		}
	}
	
	result := make([]int,0)
	for _,c := range(possibleValues) {
		if (c < 22) {
			result = append(result, c)
		}
	}
	
	if len(result) == 0 {
		return []int { getMaximum(possibleValues) }
	}
	
	return result
}

//////////////////////////////////////////////////////////////////////////////////

////////////////////////////// DECK //////////////////////////////////////////////

//struct representing the card
type Card struct {
	name string
	color string
	value int
	alternateValue int
}

//struct representing the deck
type Deck struct {
	cards []Card
}

//resets the deck
func (deck *Deck) setDeck() {
	deck.cards = make([]Card, 52)
	colors := []string { "Spades", "Hearts", "Diamonds", "Clubs" }
	names := []string { "Two", "Three", "Four", "Five", "Six", "Seven", "Eight", "Nine", "Ten", 
                            "Jack", "Queen", "King", "Ace" }                       
    count := 0
	for _, color := range(colors) {
		for _, name := range(names) {
			deck.cards[count].name = name
			deck.cards[count].color = color
			deck.cards[count].value = getValue(name)
			if name == "Ace" {
				deck.cards[count].alternateValue = 1
			} else {
				deck.cards[count].alternateValue = 0
			}
			count++
		}
	}
}

//gets a card from the deck
func (deck Deck) getCard() (Card) {
  	random := rand.Intn(52)
    return deck.cards[random]
}

/////////////////////////////////////////////////////////////////////

//used package variables
var (
	_player = BettingPlayer{}
	_dealer = Player{}
	_bet = 0
	_deck = Deck{}
)

//used contants
const (
	INGAME = 1
	STAND = 2
	BUST = 3
)


//starts the game
func run() {
	anotherRound := true
	_player.money = 100
	for anotherRound {
		bet()
		playRound()
		anotherRound = readContinue()
	}
	fmt.Println("----------------------------------------------------------------------")
	fmt.Println("Your money: $", _player.money)
	fmt.Println("----------------------------------------------------------------------")
}

//reads from the user if one wants to play another round
func readContinue() (bool) {
	if _player.money < 1 {
		fmt.Println("You've lost all your money")
		return false
	}
	
	fmt.Println("Would you like to play another round? Y/N")
	s := readFromStdInput()
    
    if s == "Y" || s == "y" {
    	return true
    }
    return false
}

//the betting operation
func bet() {
	for {
		fmt.Println("How much would you like to bet?")
		fmt.Println("Current money $", _player.money)
		bet := 0
		for bet < 1 {
			s := readFromStdInput()
			bet, _ = strconv.Atoi(s)
		}
	    
	    if ( bet > _player.money) {
	    	fmt.Println("You don't have enough money!")
	    	continue
	    }
	    
	    _bet = bet
	    _player.money -= bet
	    break	    	
    }    
}

//starts a new round of blackjack
func playRound() {
	_player.state = INGAME
	_player.cards = make([]Card,2)
	_dealer.cards = make([]Card,2)
	
	(&_deck).setDeck()
	_dealer.cards[0] = _deck.getCard()
	_dealer.cards[1] = _deck.getCard()
	_player.cards[0] = _deck.getCard()
	_player.cards[1] = _deck.getCard()
	fmt.Println("Dealer card: ", _dealer.cards[0].name, " of ",  _dealer.cards[0].color)
	for _player.state != STAND && _player.state != BUST {
		fmt.Println("----------------------------------------------------------------------") 
		fmt.Print("Your cards: ") 
		values := _player.writeCards()
		fmt.Println("----------------------------------------------------------------------")
		if(len(values) == 1 && values[0] > 21) {
			_player.state = BUST
		} else {
			answer := ""
			for answer != "h" && answer != "H" && answer != "s" && answer != "S" {
				fmt.Println("(H)it/(S)tand")
				answer = readFromStdInput()
			}
			
			if answer == "S" || answer == "s" {
				(&_player).stand(values)
		    	
		    } else {		   
		    	(&_player).hit()
		    }
		}
	}
	
	dealerTurn()
	
	if(_player.state == BUST) {
		fmt.Println("You lost")
	} else {
		if _dealer.state == BUST {
			_player.money += _bet * 2
			fmt.Println("You won $", _bet)
		} else {
			if _dealer.optimalValue < _player.optimalValue {
				_player.money += _bet * 2
				fmt.Println("You won $", _bet)
			} else if _dealer.optimalValue > _player.optimalValue {
				fmt.Println("You lost")				
			} else {
				fmt.Println("Push")
				_player.money += _bet
			}
		}
	}
	_bet = 0
}

//dealer turn operation
func dealerTurn() {
	_dealer.optimalValue = getMaximum(_dealer.calculatePossibleValues())
	_dealer.state = INGAME
	for _dealer.optimalValue < 17 {
		_dealer.hit()
		_dealer.optimalValue = getMaximum(_dealer.calculatePossibleValues())
	}
	
	if _dealer.optimalValue > 21 {
		_dealer.state = BUST
	} else {
		_dealer.state = STAND
	}
	
	fmt.Println("Dealer's cards:")
	_dealer.writeCards()
}

//gets the maximum value of the array
func getMaximum(array []int) (int) {
	v := array[0]
	i := 1
	for i < len(array) {
		if array[i] > v {
			v = array[i]
		}
		i++
	}
	return v
}

//reads a string from std input
func readFromStdInput() (string) {
	in := bufio.NewReader(os.Stdin)
	input := ""
	for {
		read, err := in.ReadString('\n')
		if err == nil {
			input = strings.Trim(read, " \r\n")
			break
		}
	}
	return input
}

//gets the maximum value of the card
func getValue(name string) (int) {
	switch name {
		case "Ace":
			return 11
		case "Two":
			return 2
		case "Three":
			return 3
		case "Four":
			return 4
		case "Five":
			return 5
		case "Six":
			return 6
		case "Seven":
			return 7
		case "Eight":
			return 8
		case "Nine":
			return 9
		case "Ten":
			return 10			
		case "Jack":
			return 10
		case "King":
			return 10
		case "Queen":
			return 10
	}
	return 0
}